#ifndef __TREF_COUNTED_EVICTOR_BASE__
#define __TREF_COUNTED_EVICTOR_BASE__

#include <string>
#include <strstream>
#include <vector>
#include <list>

#define DEBUG_TRACE

#ifdef DEBUG_TRACE
#include <iostream>
#define DEBUG_ONLY(a) a
#else
#define DEBUG_ONLY(a) 
#endif

#ifndef __CORBA_H_EXTERNAL_GUARD__
#include <omniORB3/CORBA.h>
#endif

#include <omnithread.h>

#pragma warning(disable : 4355)  // 'this' : used in base member initializer list
#pragma warning(disable : 4786)  // identifier was truncated to '255' characters in the debug information

template<class T, class T_ptr>
class TRefCountedEvictorBase;

template<class POA_Class, class T, class T_ptr> 
class TRefCountedEvictor : 
   public virtual POA_Class,
   public TRefCountedEvictorBase<T, T_ptr>
{
   public:

      typedef TRefCountedEvictor<POA_Class, T, T_ptr> RefCountedBase;

      // Our main purpose in life is to remove the ambiguity between our POA class's 
      // _this() and the ref counted base _this()...

      T_ptr _this() 
      {
         return TRefCountedEvictorBase<T, T_ptr>::_this();
      }

   	virtual void AddRef()
      {
         TRefCountedEvictorBase<T, T_ptr>::AddRef();
      }
		
      virtual void Release()
      {
         TRefCountedEvictorBase<T, T_ptr>::Release();
      }

   protected :

		TRefCountedEvictor(   
         const std::string &objectName)
         : TRefCountedEvictorBase<T, T_ptr>(objectName)
      {
      }
};

template<class T, class T_ptr>
class TRefCountedEvictorBase : 
	public PortableServer::RefCountServantBase
{
	public:
		

		virtual ~TRefCountedEvictorBase();

      T_ptr _this() 
      {
         return m_obj;
      }
      
		// reference counting

   	virtual void AddRef();
		virtual void Release();

		class ServantLocator : 
			public virtual POA_PortableServer::ServantLocator
		{
			public :

				ServantLocator()
				{
				}

				virtual PortableServer::Servant preinvoke(
					const PortableServer::ObjectId &oid,
					PortableServer::POA_ptr poa,
					const char *operation,
					PortableServer::ServantLocator::Cookie &the_cookie)
				{
					return GetObjectByReference(oid);
				}

				virtual void postinvoke(
					const PortableServer::ObjectId &oid, 
					PortableServer::POA_ptr poa, 
					const char* operation, 
					PortableServer::ServantLocator::Cookie the_cookie, 
					PortableServer::Servant the_servant)
				{
					the_servant->_remove_ref();
				}

				PortableServer::ObjectId_var AddToLocator(
					PortableServer::RefCountServantBase *pNewObject)
				{
               omni_mutex_lock locker(m_mutex);

               size_t index = 0;

               if (m_freeList.empty())
               {
   					m_servantArray.push_back(0);
					
                  index = m_servantArray.size() - 1;
               }
               else
               {
                  index = m_freeList.front();
                  m_freeList.pop_front();
               }

               m_servantArray[index] = m_lruList.NewNode(pNewObject);

               DEBUG_ONLY(std::cout << ::GetCurrentThreadId() << ": AddToLocator(" << pNewObject << ") = " << index << endl);

               return IndexToOid(index);
				}

            void RemoveFromLocator(PortableServer::ObjectId_var oid)
            {
               const size_t index = OidToIndex(oid);

               DEBUG_ONLY(std::cout << ::GetCurrentThreadId() << ": RemoveFromLocator(" << index << ")" << std::endl);

               omni_mutex_lock locker(m_mutex);

               m_lruList.RemoveNode(m_servantArray[index]);

               m_servantArray[index] = 0;

               m_freeList.push_back(index);
            }

			private :

				PortableServer::RefCountServantBase *GetObjectByReference(
					const PortableServer::ObjectId &oid)
				{
               const size_t index = OidToIndex(oid);

               omni_mutex_lock locker(m_mutex);

               LRUList::Node *pNode = m_servantArray[index];

               PortableServer::RefCountServantBase *pObj = m_lruList.UseServant(pNode);

               DEBUG_ONLY(std::cout << ::GetCurrentThreadId() << ": GetObjectByReference(" << index << ") = " << pObj << std::endl);

               m_lruList.DisplayList();

					pObj->_add_ref();

					return pObj;
				}

            size_t OidToIndex(const PortableServer::ObjectId &oid)
            {
					CORBA::String_var oid_string = PortableServer::ObjectId_to_string(oid);

               size_t index = atol(oid_string);

               omni_mutex_lock locker(m_mutex);

               if (index > m_servantArray.size() || index == 0)
               {
                  throw CORBA::OBJECT_NOT_EXIST();
               }

               index = index - 1;

               return index;
            }

            PortableServer::ObjectId_var IndexToOid(const size_t index)
            {
               std::strstream stream;

               stream << (index + 1) << std::ends;

               std::string result = stream.str();
   
               stream.rdbuf()->freeze(false);

               return PortableServer::string_to_ObjectId(result.c_str());
            }

            class LRUList
            {
               public :
                  
                  class Node
                  {
                     public :

                        Node(PortableServer::RefCountServantBase *pServant)
                           : m_pServant(pServant),
                              m_pPrev(0),
                              m_pNext(0)
                        {
                        }

                        ~Node()
                        {
                           m_pServant = 0;
                        }

                        PortableServer::RefCountServantBase *m_pServant;

                        Node *m_pPrev;
                        Node *m_pNext;
                  };

                  LRUList()
                     : m_pHead(0), m_pTail(0)
                  {

                  }

                  ~LRUList()
                  {
                     // delete them all
                  }

                  Node *NewNode(PortableServer::RefCountServantBase *pServant)
                  {
                     Node *pNode = new Node(pServant);

                     if (!m_pHead)
                     {
                        m_pHead = pNode;
                        m_pTail = pNode;
                     }
                     else
                     {
                        m_pTail->m_pNext = pNode;
                        pNode->m_pPrev = m_pTail;

                        m_pTail = pNode;
                     }
                     

                     return pNode;
                  }

                  void RemoveNode(Node *pNode)
                  {
                     UnlinkNode(pNode);

                     delete pNode;
                  }

                  PortableServer::RefCountServantBase *UseServant(Node *pNode)
                  {
                     if (m_pTail != pNode)
                     {
                        UnlinkNode(pNode);

                        // Make the node the tail

                        pNode->m_pNext = 0;
                        pNode->m_pPrev = m_pTail;
                        
                        m_pTail->m_pNext = pNode;
                        m_pTail = pNode;
                     }

                     return pNode->m_pServant;
                  }
   
                  void DisplayList()
                  {
                     Node *pNode = m_pHead;

                     std::cout << "----- List is: -----" << std::endl;

                     while (pNode)
                     {
                        std::cout << pNode << " - " << pNode->m_pServant << std::endl;

                        pNode = pNode->m_pNext;
                     }

                     std::cout << "--------------------" << std::endl;
                  }

               private :
   
                  void UnlinkNode(Node *pNode)
                  {
                     if (m_pHead == pNode)
                     {
                        m_pHead = pNode->m_pNext;
                     }
                     else if (pNode->m_pPrev)
                     {
                        pNode->m_pPrev->m_pNext = pNode->m_pNext;
                     }

                     if (m_pTail == pNode)
                     {
                        m_pTail = pNode->m_pPrev;
                     }
                     else if (pNode->m_pNext)
                     {
                        pNode->m_pNext->m_pPrev = pNode->m_pPrev;
                     }

                     pNode->m_pNext = 0;
                     pNode->m_pPrev = 0;
                  }

                  Node *m_pHead;
                  Node *m_pTail;
            };

            std::vector<LRUList::Node *> m_servantArray;
            std::list<size_t> m_freeList;

            LRUList m_lruList;

            omni_mutex m_mutex;

				// No copies
				ServantLocator(const ServantLocator &rhs);
				ServantLocator &operator=(const ServantLocator &rhs);
		};

	protected :

		TRefCountedEvictorBase(   
         const std::string &objectName);

	private :

      PortableServer::ObjectId_var m_oid;
      T_ptr m_obj;
      
      PortableServer::POA_ptr GetPOA(const std::string &objectName);

		static PortableServer::POA_ptr s_poa;
		static ServantLocator s_servantLocator;
};

template<class T, class T_ptr>
PortableServer::POA_ptr TRefCountedEvictorBase<T,T_ptr>::s_poa = 0;

template<class T, class T_ptr>
TRefCountedEvictorBase<T,T_ptr>::ServantLocator TRefCountedEvictorBase<T,T_ptr>::s_servantLocator;

template<class T, class T_ptr>
TRefCountedEvictorBase<T,T_ptr>::TRefCountedEvictorBase<T,T_ptr>(
   const std::string &objectName) 
   :  m_oid(s_servantLocator.AddToLocator(this)),
      m_obj(T::_narrow(GetPOA(objectName)->create_reference_with_id(m_oid, T::_PD_repoId)))
{ 
   DEBUG_ONLY(std::cout << "TRefCountedEvictorBase() - " << this << endl); 
}

template<class T, class T_ptr>
TRefCountedEvictorBase<T,T_ptr>::~TRefCountedEvictorBase<T,T_ptr>() 
{ 
   DEBUG_ONLY(std::cout << "~TRefCountedEvictorBase() - " << this << std::endl); 

   s_servantLocator.RemoveFromLocator(m_oid);
}

template<class T, class T_ptr>
void TRefCountedEvictorBase<T,T_ptr>::AddRef()
{
	RefCountServantBase::_add_ref();

	DEBUG_ONLY(std::cout << "TRefCountedEvictorBase(" << this << ")::AddRef()" << std::endl);
}

template<class T, class T_ptr>
void TRefCountedEvictorBase<T,T_ptr>::Release()
{
	DEBUG_ONLY(std::cout << "TRefCountedEvictorBase(" << this << ")::Release()" << std::endl);

	RefCountServantBase::_remove_ref();
}

template<class T, class T_ptr>
PortableServer::POA_ptr TRefCountedEvictorBase<T,T_ptr>::GetPOA(const std::string &objectName)
{
	if (0 == s_poa)
	{
		int argc = 0;

      CORBA::ORB_var orb = CORBA::ORB_init(argc ,0,"");

		CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
		PortableServer::POA_var rootPOA = PortableServer::POA::_narrow(obj);

		PortableServer::POAManager_var POAManager = rootPOA->the_POAManager();

		CORBA::PolicyList policyList;
		policyList.length(3);
		policyList[0] = rootPOA->create_request_processing_policy(PortableServer::USE_SERVANT_MANAGER);
		policyList[1] = rootPOA->create_servant_retention_policy(PortableServer::NON_RETAIN);
		policyList[2] = rootPOA->create_id_assignment_policy(PortableServer::USER_ID);

      std::string POAName = objectName + "::POA";

		s_poa = rootPOA->create_POA(POAName.c_str(), POAManager, policyList);

		// Activate a servant locator in the root poa.
    
		PortableServer::ObjectId_var id = rootPOA->activate_object(&s_servantLocator);
		PortableServer::ServantLocator_var slref = s_servantLocator._this();
		s_servantLocator._remove_ref();

		// Register the servant activator with our new poa.
		s_poa->set_servant_manager(slref);   
   }

	return s_poa;
}

#endif // __TREF_COUNTED_EVICTOR_BASE__
